/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#ifndef _N80_H_
#define _N80_H_
#include "../z80/z80.h"
#ifdef __cplusplus
extern "C" {
#endif
#define export
typedef const DWORD	CDWORD, * LPCDWORD;
typedef const BYTE	CBYTE, * LPCBYTE;
/**--------------------------------------------------------------------------
 **		PROTOTYPE DECLARATIONs
 **------------------------------------------------------------------------*/
struct tagCFGITEM;
typedef struct tagCFGITEM	CFGITEM, *LPCFGITEM;
struct tagCFGPAGE;
typedef struct tagCFGPAGE	CFGPAGE, *LPCFGPAGE;
struct CONFIG_tag;
typedef struct CONFIG_tag	CONFIG, *LPCONFIG;
/**--------------------------------------------------------------------------
 **		\
 **------------------------------------------------------------------------*/
typedef struct tagKEYEVT {
	USHORT  key;
	BOOL    SHIFT;
	BOOL    ALT;
	BOOL    CTL;
	char    ascii;
}   KEYEVTG, *LPKEYEVT;

struct tagCFGITEM {
	int			id;
	int			type;
	char*		name;
	char*		hint;
	int			value;
	int			(*controller)(LPCFGITEM, int, LPVOID);
	LPVOID		arg;
	LPCFGPAGE	parent;
};

struct tagCFGPAGE {
	LPCONFIG	parent;
	int			page;
	LPVOID		arg;
	LPCFGITEM	items;
	int			item_cnt;
	int			cur_idx;
	LPCFGITEM	(*get_item)(LPCFGPAGE);
	int			(*handler)(LPCFGPAGE, LPKEYEVT);
	void		(*prepare)(LPCFGPAGE, LPVOID arg);
	int			(*call)(LPCFGPAGE, int, LPVOID);
	void		(*reflesh)(LPCFGPAGE);
	int			jmp_page;
	LPVOID		jmp_arg;
};

struct CONFIG_tag {
	char		text[25][80];
	char		attr[25][80];
	KEYEVTG		kv;
	CFGPAGE		pages[5];
	int			page;
	int			maxpages;
	void		(*init)(LPCONFIG);
	void		(*open)(LPCONFIG);
	void		(*init_page)(LPCONFIG, int page, LPVOID arg);
	BOOL		(*dispatch)(LPCONFIG, LPKEYEVT);
	void		(*drawitem)(LPCFGITEM item, int line, BOOL selected);
	void		(*show)(LPCONFIG);
	char		(*to_ascii)(LPKEYEVT kv);
};
struct CONFIG_arg_selector {
	const char	**items;
	int			maxlen;
	int			count;
};
typedef struct cpuid_tag {
	BOOL		avx2_ready;
	BOOL		avx1_ready;
	BOOL		sse_ready;
	BOOL		sse2_ready;
	BOOL		sse3_ready;
	BOOL		ssse3_ready;
	BOOL		sse41_ready;
	BOOL		sse42_ready;
	BOOL		mmx_ready;
	int			simd_level;				// 0..9 = N/A .. AVX2
}	CPUID, *LPCPUID;

typedef struct N80_tag {				// N80 Win32
	LPCPUID		cpuid;					// cpuid information
	BOOL		use_avx;				// use avx code to draw if possible
	BOOL		not_use_mmx;			// never use simd code to draw
	int			vtbl_vram20[20];		// DirectSurface VRAM offset table (20 Lines)
	int			vtbl_vram25[25];		// DirectSurface VRAM offset table (25 Lines) 
	LPBYTE		vram_addr;				// DirectSurface VRAM pointer
	DWORD		width;					// DirectSurface width
	DWORD		height;					// DirectSurface height
	DWORD		vram_lpitch;			// DirectSurface width pitch
	int			vram_pixcel_pitch;		// DirectSurface width pixcel pitch
	int			vram_vpitch[20];		// DirectSurface virtical pixcel pitch
	DWORD		vram_top_offset;		// DirectSurface top start line
	DWORD		vram_bottom_offset;		// DirectSurface bottom end line
	DWORD		vram_left_offset;		// DirectSurface left start width
	DWORD		vram_right_offset;		// DirectSurface right end width
	DWORD		screen_vpadding;		// screen top/bottom padding pixcels
	DWORD		screen_hpadding;		// screen left/right padding pixcels
	LPCBYTE		prime_palettes;			// primary palettes
	LPCBYTE		second_palettes;		// secondary paletts (interlace)
	LPCDWORD	prime_rgb_colors;		// primary rgb colors
	LPCDWORD	second_rgb_colors;		// seconday rgb colors (interlace)
	DWORD		cur_pix_rgb;			// current font rgb color
	DWORD		cur_pix_rgb2;			// current font rgb color (interlace)
	BYTE		cur_pix_rgb_pal;		// current font color palette no
	BYTE		cur_pix_rgb_pal2;		// current font color palette no (interlace)
	LONG		hard_volume;			// default DirectSound Hardware Volume
	LONG		soft_volume;			// default DirectSound Software Volume
	BOOL		tenkeyless;				// use ten keyless keyboard
	SHORT		screen_size;			// Window(Full-Screen) size (0:autoA1:720x480)
	BOOL		m_z80_no_wait;			// z80 no-wait mode
	TCHAR		n80_rom_file_name[64];	// ROM file names
	TCHAR		_4th_rom_file_name[64];
	TCHAR		fdu_rom_file_name[64];

	int			tmp_mix_mode;
	DWORD		tmp_save_screen_mode;
	BOOL		tmp_save_use_new_pcg;
	BYTE		tmp_save_text_visible;
	CONFIG		config;

	double		eff_hz;					// Z80Clock(Hz)
	double		sys_hz;					// Z80_Clock(Hz)
	int			flips;					// bԕ`
	double		lines_per_flip;			// ϕ`֍s

	double		remain_ticks_top_sum;
	double		remain_ticks_sum;
	double		remain_minus_sum;
	LONGLONG	remain_top_minus_cnt;
	LONGLONG	remain_minus_cnt;
	LONGLONG	loop_cnt;
	TCHAR		sys_info[512];

	LONG		(*get_hard_volume)(void);
	void		(*set_hard_volume)(LONG);
	SHORT		(*get_soft_volume)(void);
	void		(*set_soft_volume)(SHORT);
	void		(*set_tenkeyless)(BOOL);
	void		(*save_blt_context)(void);
	void		(*restore_blt_context)(void);

}	N80, *LPN80;
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern N80	n80win32;

#ifdef __cplusplus
}	// extern "C"
#endif
#endif//_N80_H_
