/**--------------------------------------------------------------------------
 **
 **		NEC PC-8001 Emulator "N80" for Win32/DirectX
 **		--------------------------------------------
 **		Copyright (c) 1993-2010 H.Kanemoto
 **
 **		Preference
 **
 **		Rev		Date		Descriptions
 **		-----	----------	-----------------------------------------------
 **		1.00	1996/06/27	porting DOS version on WinG
 **		1.20	1997/08/10	porting WinG version on Win32/DirectX
 **		1.30	1998/03/12	
 **		1.43	2007/09/01	XPΉ
 **		1.44	2009/12/15	Vista/7Ή
 **		1.51	2010/01/27	DirectSound OnlyɕύX
 **------------------------------------------------------------------------*/
#include "n80_form.h"
#include "n80_directx.h"
#include "n80_assay.h"
#include "../z80/z80_thread.h"
#include "../res/resource.h"
#include <stdio.h>
#include <commctrl.h>
/**--------------------------------------------------------------------------
 **		O[oϐ
 **------------------------------------------------------------------------*/
extern n80_form		*form;
extern z80_thread	*th;
extern n80_directx	*dx;
/**--------------------------------------------------------------------------
 **	declare macro
 **------------------------------------------------------------------------*/
#define TYPE_TOGGLE     (0)     // toggle control
#define TYPE_TEXTIN     (1)     // text-in control
#define TYPE_UPDOWN     (2)     // value up-down control
#define TYPE_LINK       (3)     // link to other page
#define TYPE_SELECT     (4)     // select item control
#define TYPE_DRVSELECT  (100)   // drive selector
#define TYPE_IMGSLOT    (101)   // disk image slot
#define TYPE_TERMINATOR (-1)    // terminator dummy control
#define VERB_INIT       (0)     // INIT control
#define VERB_EVENT      (1)     // Key Event handle
#define VERB_PRINT      (2)     // Print control
#define VERB_PRINT_SEL  (3)     // Print control (selected status)
#define VERB_COMMIT		(4)     // on toggled set value
#define VERB_TEXTIN_CH  (5)     // on key char
#define ACT_NONE        (0)
#define ACT_DONE        (1)
#define ACT_CHANGEPAGE  (9)
#define ACT_QUIT        (-1)
#define CHAR_VALID      (1)
#define CHAR_INVALID    (0)
#define SP              (32)
 /**--------------------------------------------------------------------------
  **	local function declaration
  **------------------------------------------------------------------------*/
static int _controller_crt(LPCFGITEM, int, LPVOID);
static int _controller_use_newpcg(LPCFGITEM, int, LPVOID);
static int _controller_tenkeyless(LPCFGITEM, int, LPVOID);
static int _controller_interrace(LPCFGITEM, int, LPVOID);
static int _controller_sync_vb(LPCFGITEM, int, LPVOID);
static int _controller_nowait(LPCFGITEM, int, LPVOID);
static int _controller_beepmaker(LPCFGITEM, int, LPVOID);
static int _controller_useavx(LPCFGITEM, int, LPVOID);
static int _controller_notsimd(LPCFGITEM, int, LPVOID);
static int _controller_n80rom(LPCFGITEM, int, LPVOID);
static int _controller_4throm(LPCFGITEM, int, LPVOID);
static int _controller_fdurom(LPCFGITEM, int, LPVOID);
static int _controller_db(LPCFGITEM, int, LPVOID);
static int _controller_ap(LPCFGITEM, int, LPVOID);
static int _controller_image_slot(LPCFGITEM, int, LPVOID);
static int _controller_drive_select(LPCFGITEM, int, LPVOID);
static int _controller_common_link(LPCFGITEM, int, LPVOID);
static int _controller_selector(LPCFGITEM self, int verb, LPVOID param);
#if _DEBUG
static int _controller_debug_main(LPCFGITEM, int, LPVOID);
static int _controller_debug_sub(LPCFGITEM, int, LPVOID);
#endif
/**--------------------------------------------------------------------------
 **	local variables
 **------------------------------------------------------------------------*/
static const char *mmxitems[] = {"------", "MMX   ", "SSE   ", "SSE2  ", "SSE3  ", "SSSE3 ", "SSE4.1", "SSE4.2", "AVX1  ", "AVX2  " };
static CONFIG_arg_selector mmxarg = {mmxitems, 6, 10};
static CFGITEM page0_items[] = {
	{0xF3, TYPE_LINK,   "ROM File name settings ............... [\x1c]", "Link to ROM File names setting page", 2, _controller_common_link, NULL},
	{0xF4, TYPE_LINK,   "misc settings ........................ [\x1c]", "Link to miscellaneous settings page", 3, _controller_common_link, NULL},
#if _DEBUG
	{0xF5, TYPE_LINK,   "debug settings ....................... [\x1c]", "Link to z80 debug settings page", 4, _controller_common_link, NULL},
#endif
	{0x04, TYPE_TOGGLE, "TEN-Keyless .......................... [ ]", "assign Ten-Key (2/4/6/8) to cursor key for Tenkey-less keyboard", 0, _controller_tenkeyless, NULL},
	{0x05, TYPE_TOGGLE, "Interrace-like effect ................ [ ]", "display like interrace monitor", 0, _controller_interrace, NULL},
	{0x21, TYPE_UPDOWN, "Volume (HARD)  ....................... [    ](dB)", "Sound Hardware Volume", 0, _controller_db, NULL},
	{0x21, TYPE_UPDOWN, "Amplitude (SOFT)  .................... [  ]", "Sound Software Volume", 0, _controller_ap, NULL},
	{0x41, TYPE_IMGSLOT,"Image File Slot #0 Control ........... [ ]", "[DEL] remove [INS] load [RET] link to insert/eject disk image page", 0, _controller_image_slot},
	{0x42, TYPE_IMGSLOT,"Image File Slot #1 Control ........... [ ]", "[DEL] remove [INS] load [RET] link to insert/eject disk image page", 0, _controller_image_slot},
	{0x00, TYPE_TERMINATOR}
};
static CFGITEM page1_items[] = {
	{0x01, TYPE_DRVSELECT, " 0) .................................. [ ]", "inser/eject disk image #0 to/from drive [A or B]", -1, _controller_drive_select},
	{0x02, TYPE_DRVSELECT, " 1) .................................. [ ]", "inser/eject disk image #1 to/from drive [A or B]", -1, _controller_drive_select},
	{0x03, TYPE_DRVSELECT, " 2) .................................. [ ]", "inser/eject disk image #2 to/from drive [A or B]", -1, _controller_drive_select},
	{0x04, TYPE_DRVSELECT, " 3) .................................. [ ]", "inser/eject disk image #3 to/from drive [A or B]", -1, _controller_drive_select},
	{0x05, TYPE_DRVSELECT, " 4) .................................. [ ]", "inser/eject disk image #4 to/from drive [A or B]", -1, _controller_drive_select},
	{0x06, TYPE_DRVSELECT, " 5) .................................. [ ]", "inser/eject disk image #5 to/from drive [A or B]", -1, _controller_drive_select},
	{0x07, TYPE_DRVSELECT, " 6) .................................. [ ]", "inser/eject disk image #6 to/from drive [A or B]", -1, _controller_drive_select},
	{0x08, TYPE_DRVSELECT, " 7) .................................. [ ]", "inser/eject disk image #7 to/from drive [A or B]", -1, _controller_drive_select},
	{0x09, TYPE_DRVSELECT, " 8) .................................. [ ]", "inser/eject disk image #8 to/from drive [A or B]", -1, _controller_drive_select},
	{0x0A, TYPE_DRVSELECT, " 9) .................................. [ ]", "inser/eject disk image #9 to/from drive [A or B]", -1, _controller_drive_select},
};
static CFGITEM page2_items[] = {
	{0x01, TYPE_TEXTIN, "N80 rom file name .................... [                        ]", "N-BASIC ROM file name (required)", 20, _controller_n80rom},
	{0x02, TYPE_TEXTIN, "4th rom file name .................... [                        ]", "4TH ROM file name (optional)", 20, _controller_4throm},
	{0x03, TYPE_TEXTIN, "Disk Unit rom file name .............. [                        ]", "Disk ROM file name (optional)", 20, _controller_fdurom},
	{0x00, TYPE_TERMINATOR}
};
static CFGITEM page3_items[] = {
	{0x01, TYPE_SELECT,	"CPU SIMD capability .................. [      ]", "cpu suport mmx/sse/avx level", 0, _controller_selector, &mmxarg},
	{0x02, TYPE_TOGGLE, "Not use SIMD function to draw ........ [ ]", "do not use SIMD code to draw screen. If set ignore below setting.", 0, _controller_notsimd, NULL},
	{0x03, TYPE_TOGGLE, "Use AVX function to draw ............. [ ]", "use AVX code to draw screen if possible", 0, _controller_useavx, NULL},
	//{0x04, TYPE_TOGGLE, "HiRes-CRT (24kHz/400Lines)............ [ ]", "select CRT from 24kHz or 15kHz(Normal/200Lines)", 0, _controller_crt, NULL},
	//{0x04, TYPE_TOGGLE, "No-wait .............................. [ ]", "ignore memory wait, M1 wait only", 0, _controller_nowait, NULL},
	{0x06, TYPE_TOGGLE, "New PCG (chiqlappe) .................. [ ]", "use new pcg (chiqlappe) device", 0, _controller_use_newpcg, NULL},
	{0x07, TYPE_TOGGLE, "sync Vertical-blank .................. [ ]", "draw in synchronization with the vertical retrace signal", 0, _controller_sync_vb, NULL},
	{0x08, TYPE_TOGGLE, "Beep<=>PCM converter ................. [ ]", "use legacy pcm converter function", 0, _controller_beepmaker, NULL},
	{0x00, TYPE_TERMINATOR}
};
static CFGITEM page4_items[] = {
#if _DEBUG
	{0x01, TYPE_TOGGLE, "Debug Main CPU ....................... [ ]", "toggle main cpu debugger", 0, _controller_debug_main, NULL},
	{0x02, TYPE_TOGGLE, "Debug Sub CPU ........................ [ ]", "toggle sub cpu debugger", 0, _controller_debug_sub, NULL},
#endif
	{0x00, TYPE_TERMINATOR}
};
static LPCFGITEM page_items[] = { page0_items, page1_items, page2_items, page3_items, page4_items, NULL };
#pragma region Windows Virtual Key<=>ASCII Code
static char to_ascii(LPKEYEVT kv) {
	switch (kv->key) {
	case 0x31:	/* 1 */	return kv->SHIFT ? 0x21 : 0x31;
	case 0x32:	/* 2 */	return kv->SHIFT ? 0x22 : 0x32;
	case 0x33:	/* 3 */	return kv->SHIFT ? 0x23 : 0x33;
	case 0x34:	/* 4 */	return kv->SHIFT ? 0x24 : 0x34;
	case 0x35:	/* 5 */	return kv->SHIFT ? 0x24 : 0x35;
	case 0x36:	/* 6 */	return kv->SHIFT ? 0x26 : 0x36;
	case 0x37:	/* 7 */	return kv->SHIFT ? 0x27 : 0x37;
	case 0x38:	/* 8 */	return kv->SHIFT ? 0x28 : 0x38;
	case 0x39:	/* 9 */	return kv->SHIFT ? 0x29 : 0x39;
	case 0x30:	/* 0 */	return kv->SHIFT ? 0x30 : 0x30;

	case 0x51:	/* Q */	return kv->SHIFT ? 0x51 : 0x71;
	case 0x57:	/* W */	return kv->SHIFT ? 0x57 : 0x77;
	case 0x45:	/* E */	return kv->SHIFT ? 0x45 : 0x65;
	case 0x52:	/* R */	return kv->SHIFT ? 0x52 : 0x72;
	case 0x54:	/* T */	return kv->SHIFT ? 0x54 : 0x74;
	case 0x59:	/* Y */	return kv->SHIFT ? 0x59 : 0x79;
	case 0x55:	/* U */	return kv->SHIFT ? 0x55 : 0x75;
	case 0x49:	/* I */	return kv->SHIFT ? 0x49 : 0x69;
	case 0x4F:	/* O */	return kv->SHIFT ? 0x4F : 0x6F;
	case 0x50:	/* P */	return kv->SHIFT ? 0x50 : 0x70;

	case 0x41:	/* A */	return kv->SHIFT ? 0x41 : 0x61;
	case 0x53:	/* S */	return kv->SHIFT ? 0x53 : 0x73;
	case 0x44:	/* D */	return kv->SHIFT ? 0x44 : 0x64;
	case 0x46:	/* F */	return kv->SHIFT ? 0x46 : 0x66;
	case 0x47:	/* G */	return kv->SHIFT ? 0x47 : 0x67;
	case 0x48:	/* H */	return kv->SHIFT ? 0x48 : 0x68;
	case 0x4A:	/* J */	return kv->SHIFT ? 0x4A : 0x6A;
	case 0x4B:	/* K */	return kv->SHIFT ? 0x4B : 0x6B;
	case 0x4C:	/* L */	return kv->SHIFT ? 0x4C : 0x6C;

	case 0x5A:	/* Z */	return kv->SHIFT ? 0x5A : 0x7A;
	case 0x58:	/* X */	return kv->SHIFT ? 0x58 : 0x78;
	case 0x43:	/* C */	return kv->SHIFT ? 0x43 : 0x63;
	case 0x56:	/* V */	return kv->SHIFT ? 0x56 : 0x76;
	case 0x42:	/* B */	return kv->SHIFT ? 0x42 : 0x62;
	case 0x4E:	/* N */	return kv->SHIFT ? 0x4E : 0x6E;
	case 0x4D:	/* M */	return kv->SHIFT ? 0x4D : 0x6D;

	case VK_ESCAPE:     /*ESC*/     return 0x1B;
	case VK_BACK:       /* BS*/     return 0x08;
	case VK_DELETE:     /*DEL*/     return 0x08;
	case VK_TAB:        /*TAB*/     return 0x09;
	case VK_RETURN:     /*RET*/     return 0x0D;

	case VK_NUMPAD0:	/* 0 */     return 0x30;
	case VK_NUMPAD1:	/* 1 */     return 0x31;
	case VK_NUMPAD2:	/* 2 */     return 0x32;
	case VK_NUMPAD3:	/* 3 */     return 0x33;
	case VK_NUMPAD4:	/* 4 */     return 0x34;
	case VK_NUMPAD5:	/* 5 */     return 0x35;
	case VK_NUMPAD6:	/* 6 */     return 0x36;
	case VK_NUMPAD7:	/* 7 */     return 0x37;
	case VK_NUMPAD8:	/* 8 */     return 0x38;
	case VK_NUMPAD9:	/* 9 */     return 0x39;
	case VK_MULTIPLY:	/* * */     return 0x2A;
	case VK_ADD:    	/* + */     return 0x2B;
	case VK_SEPARATOR:	/* , */     return 0x2C;
	case VK_SUBTRACT:	/* - */     return 0x2D;
	case VK_DECIMAL:	/* . */     return 0x2E;
	case VK_DIVIDE:	    /* / */     return 0x2F;

	case VK_OEM_MINUS:	/* = -   */	return kv->SHIFT ? 0x3D : 0x2D;
	case VK_OEM_7:	    /* ^ ~   */	return kv->SHIFT ? 0x7E : 0x5E;
	case VK_OEM_3:	    /* @ `   */	return kv->SHIFT ? 0x60 : 0x40;
	case VK_OEM_4:	    /* [ {   */	return kv->SHIFT ? 0x7B : 0x5B;
	case VK_OEM_PLUS:	/* + ;   */	return kv->SHIFT ? 0x2B : 0x3B;
	case VK_OEM_1:	    /* * :   */	return kv->SHIFT ? 0x2A : 0x3A;
	case VK_OEM_6:	    /* ] }   */	return kv->SHIFT ? 0x7D : 0x5D;
	case VK_OEM_COMMA:  /* < ,   */	return kv->SHIFT ? 0x3C : 0x2C;
	case VK_OEM_PERIOD: /* > .   */	return kv->SHIFT ? 0x3E : 0x2E;
	case VK_OEM_2:	    /* / ?   */	return kv->SHIFT ? 0x3F : 0x2F;
	case VK_OEM_102:	/* \ _   */	return kv->SHIFT ? 0x5F : 0x5C;
	case VK_OEM_5:      /* \ |   */	return kv->SHIFT ? 0x7C : 0x5C;
	default:    return '\0';
	}
}
#pragma endregion
/**--------------------------------------------------------------------------
 **	functions
 **------------------------------------------------------------------------*/
static bool is_alpha(LPKEYEVT kv) {
	if (kv->ascii >= 'a' && kv->ascii <= 'z')
		return true;
	if (kv->ascii >= 'A' && kv->ascii <= 'Z')
		return true;
	return false;
}
static bool is_numeric(LPKEYEVT kv) {
	return (kv->ascii >= '0' && kv->ascii <= '9');
}
#pragma region Toggle Controller
static void _toggle_print(LPCFGITEM self, char* vram, char on) {
	memset(vram, SP, 78);
	strcpy(vram, self->name);
	if (self->value == 1)
		vram[40] = on;
}
static int _controller_common_toggle(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_EVENT: {
		LPKEYEVT kv = (LPKEYEVT)param;
		if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
			self->value = self->value == 1 ? 0 : 1;
			self->controller(self, VERB_COMMIT, NULL);
			return ACT_DONE;
			} else
			return ACT_NONE;
	}
	case VERB_PRINT:
	case VERB_PRINT_SEL:
		_toggle_print(self, (char*)param, 'X');
		break;
	}
	return ACT_NONE;
}
static int _controller_crt(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80device.vrtc.crt_hires_mode == _CRT_HIRES ? 1 : 0;				break;
	case VERB_COMMIT:
		n80device.vrtc.crt_hires_mode = self->value == 1 ? _CRT_HIRES : _CRT_NORMAL;	break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_use_newpcg(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80win32.tmp_save_use_new_pcg ? 1 : 0;	break;
	case VERB_COMMIT:
		n80win32.tmp_save_use_new_pcg = self->value == 1;		break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_beepmaker(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = dx->getBeepMakerType();		break;
	case VERB_COMMIT:
		dx->setBeepMakerType((short)self->value);	break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_useavx(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80win32.use_avx ? 1 : 0;		break;
	case VERB_COMMIT:
		n80win32.use_avx = self->value == 1;
		n80device.blt.set_simd_level(n80win32.cpuid->simd_level);
		break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_notsimd(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80win32.not_use_mmx ? 1 : 0;		break;
	case VERB_COMMIT:
		n80win32.not_use_mmx = self->value == 1;
		n80device.blt.set_simd_level(n80win32.cpuid->simd_level);
		break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_sync_vb(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = dx->getSyncVBlank() ? 1 : 0;	break;
	case VERB_COMMIT:
		dx->setSyncVBlank(self->value == 1);		break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_tenkeyless(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80win32.tenkeyless ? 1 : 0;	break;
	case VERB_COMMIT:
		n80win32.set_tenkeyless(self->value == 1);	break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_interrace(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80device.param.interlace ? 1 : 0;		break;
	case VERB_COMMIT:
		n80device.param.interlace = self->value == 1;
		n80device.blt.chg_palette(n80device.param.interlace);	break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_nowait(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80win32.m_z80_no_wait ? 1 : 0;	break;
	case VERB_COMMIT:
		n80win32.m_z80_no_wait = self->value == 1;		break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
#if _DEBUG
static int _controller_debug_main(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80device.cpu.dbgctx.enable;	break;
	case VERB_COMMIT:
		n80device.cpu.dbgctx.debug_enable(&n80device.cpu.dbgctx, (BYTE)self->value);
		break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
static int _controller_debug_sub(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = n80device.fdu.cpu.dbgctx.enable ? 1 : 0;
	case VERB_COMMIT:
		n80device.fdu.cpu.dbgctx.debug_enable(&n80device.fdu.cpu.dbgctx, (BYTE)self->value);
		break;
	default:
		return _controller_common_toggle(self, verb, param);
	}
	return ACT_NONE;
}
#endif
#pragma endregion
#pragma region Link Controller
static int _controller_common_link(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_EVENT: {
		LPKEYEVT kv = (LPKEYEVT)param;
		if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
			self->parent->jmp_arg = self->arg;
			self->parent->jmp_page = self->value;
			return ACT_CHANGEPAGE;
		}
		break;
	}
	case VERB_PRINT:
	case VERB_PRINT_SEL: {
		char* vram = (char*)param;
		memset(vram, SP, 78);
		strcpy(vram, self->name);
		break;
	}
	}
	return ACT_NONE;
}
#pragma endregion
#pragma region UpDn Controller
static int _controller_common_updown(LPCFGITEM self, int verb, LPVOID param, int minvalue, int maxvalue, int maxlen) {
	switch (verb) {
	case VERB_EVENT: {
		LPKEYEVT kv = (LPKEYEVT)param;
		if (kv->key == VK_LEFT) {
			if (self->value > minvalue) {
				--self->value;
				self->controller(self, VERB_COMMIT, NULL);
				return ACT_DONE;
			}
		} else if (kv->key == VK_RIGHT) {
			if (self->value < maxvalue) {
				++self->value;
				self->controller(self, VERB_COMMIT, NULL);
				return ACT_DONE;
			}
		}
		break;
	}
	case VERB_PRINT:
	case VERB_PRINT_SEL: {
		char* vram = (char*)param;
		char tmp[64];
		memset(vram, SP, 78);
		strcpy(vram, self->name);
		memset(tmp, 0, 64);
		sprintf(tmp, "%d", self->value);
		char* ptr = &vram[40];
		int len = strlen(tmp);
		if (len < maxlen) {
			for (int i = 0; i < (maxlen - len); ++i)
				*ptr++ = SP;
		}
		strcpy(ptr, tmp);
		vram[40 + maxlen] = ']';
		break;
	}
	}
	return ACT_NONE;
}
static int _controller_db(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = (short)(n80win32.get_hard_volume() - 1);	break;
	case VERB_COMMIT:
		n80win32.set_hard_volume((short)self->value);			break;
	default:
		return _controller_common_updown(self, verb, param, -100, 0, 4);
	}
	return ACT_NONE;
}
static int _controller_ap(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = (short)n80win32.get_soft_volume();		break;
	case VERB_COMMIT:
		n80win32.set_soft_volume((short)self->value);			break;
	default:
		return _controller_common_updown(self, verb, param, 1, 18, 2);
	}
	return ACT_NONE;
}
#pragma endregion
#pragma region Text Controller
static int _controller_common_filename(LPCFGITEM self, int verb, LPVOID param, char* buff, const char* defval) {
	switch (verb) {
	case VERB_EVENT: {
		LPKEYEVT kv = (LPKEYEVT)param;
		if (kv->key == VK_RETURN) {
			self->controller(self, VERB_COMMIT, kv);
			return ACT_DONE;
		}
		int len = strlen(buff);
		if (len > self->value)
			return ACT_NONE;
		if (kv->key == VK_BACK && len > 0) {    // [BS] Key pressed
			buff[len - 1] = '\0';
			return ACT_DONE;
		} else if (len < self->value && self->controller(self, VERB_TEXTIN_CH, kv) == CHAR_VALID) {
			buff[len] = kv->ascii;
			buff[len + 1] = '\0';
			return ACT_DONE;
		}
		return ACT_NONE;
	}
	case VERB_PRINT:
	case VERB_PRINT_SEL: {
		char* vram = (char*)param;
		int l = strlen(buff);
		memset(vram, SP, 78);
		strcpy(vram, self->name);
		for (int i = 0; i < self->value; ++i) {
			char ch = *buff++;
			if (ch == '\0')
				break;
			vram[40 + i] = ch;
		}
		// show guide cursor when selected
		if (verb == VERB_PRINT_SEL) {
			if (l < self->value)
				vram[40 + l] = 0x1D;
		}
	}
	case VERB_TEXTIN_CH: {
		LPKEYEVT kv = (LPKEYEVT)param;
		if (is_alpha(kv))
			return CHAR_VALID;
		if (is_numeric(kv))
			return CHAR_VALID;
		if (kv->ascii == '_' || kv->ascii == '.')
			return CHAR_VALID;
		if (kv->ascii == '-' && strlen(buff) > 0)
			return CHAR_VALID;
		return CHAR_INVALID;
	}
	case VERB_COMMIT:
		if (strlen(buff) == 0 && defval != NULL)
			strcpy(buff, defval);
		break;
	}
	return ACT_NONE;
}
static int _controller_n80rom(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80win32.n80_rom_file_name, "8801-N80.ROM");
}
static int _controller_4throm(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80win32._4th_rom_file_name, "");
}
static int _controller_fdurom(LPCFGITEM self, int verb, LPVOID param) {
	return _controller_common_filename(self, verb, param, n80win32.fdu_rom_file_name, "");
}
#pragma endregion
#pragma region Misc Controller
static int _controller_image_slot(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = self->arg == NULL ? 0 : 1;
		break;
	case VERB_EVENT: {
		LPKEYEVT kv = (LPKEYEVT)param;
		if (self->value == 1) {
			if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
				self->parent->jmp_arg = self->arg;
				self->parent->jmp_page = 1;
				return ACT_CHANGEPAGE;
			} else if (kv->key == VK_DELETE) {
				int slot_no = (self->id & 0x0f) - 1;
				n80device.fdu.mgr.detach(&n80device.fdu.mgr, slot_no);
				self->arg = NULL;
				self->value = 0;
				return ACT_DONE;
			}
		} else if (kv->key == VK_INSERT) {
			int slot_no = (self->id & 0x0f) - 1;
			form->chooser(slot_no);
			self->parent->jmp_arg = NULL;
			self->parent->jmp_page = 0;
			return ACT_CHANGEPAGE;
		}
		return ACT_NONE;
	}
	case VERB_PRINT:
	case VERB_PRINT_SEL:
		_toggle_print(self, (char*)param, '\x1c');
		break;
	case VERB_COMMIT:
		break;
	}
	return ACT_NONE;
}
static int _controller_drive_select(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
	case VERB_INIT:
		self->value = ((LPFLOPPYDISK)self->arg)->unit_no;
		break;
	case VERB_EVENT: {
		LPKEYEVT kv = (LPKEYEVT)param;
		if (kv->key == VK_SPACE || kv->key == VK_RETURN) {
			self->controller(self, VERB_COMMIT, NULL);
			return ACT_CHANGEPAGE;
		} else if (kv->key == VK_LEFT) {
			switch (self->value) {
			case 0:	self->value = -1;	break;
			case 1: self->value = 0;	break;
			default:self->value = 1;	break;
			}
			return ACT_DONE;
		} else if (kv->key == VK_RIGHT) {
			switch (self->value) {
			case 0:	self->value = 1;	break;
			case 1: self->value = -1;	break;
			default:self->value = 0;	break;
			}
			return ACT_DONE;
		}
		return ACT_NONE;
	}
	case VERB_PRINT:
	case VERB_PRINT_SEL: {
		LPFLOPPYDISK disk = (LPFLOPPYDISK)self->arg;
		char* vram = (char*)param;
		memset(vram, SP, 78);
		strcpy(vram, self->name);
		// copy disk name
		char* title = disk->header->title[0] == 0 ? "{NO TITLE}\0\0\0\0\0\0\0" : disk->header->title;
		for (int j = 0; j < 17; ++j) {
			char c = title[j];
			vram[j + 4] = c == '\0' ? '.' : c;
		}
		// mount status
		char c = ' ';
		switch (self->value) {
		case 0: c = 'A'; break;
		case 1: c = 'B'; break;
		}
		vram[40] = c;
		break;
	}
	case VERB_COMMIT: {
		LPFLOPPYDISK disk = (LPFLOPPYDISK)self->arg;
		if (disk->unit_no != self->value) {
			if (disk->unit_no == 0 || disk->unit_no == 1)
				n80device.fdu.fdd.eject(&n80device.fdu.fdd, disk->unit_no);
			if (self->value == 0 || self->value == 1) {
				n80device.fdu.fdd.eject(&n80device.fdu.fdd, self->value);
				n80device.fdu.fdd.insert(&n80device.fdu.fdd, self->value, disk);
			}
			self->value = disk->unit_no;
			self->parent->reflesh(self->parent);
		}
		break;
	}
	}
	return ACT_NONE;
}
static int _controller_selector(LPCFGITEM self, int verb, LPVOID param) {
	switch (verb) {
		case VERB_INIT:
			self->value = n80win32.cpuid->simd_level;
			break;
		case VERB_PRINT:
		case VERB_PRINT_SEL: {
			CONFIG_arg_selector *arg = (CONFIG_arg_selector *)self->arg;
			char* vram = (char*)param;
			memset(vram, SP, 78);
			strcpy(vram, self->name);
			vram = &vram[40];
			for (int i = 0; i < arg->maxlen; ++i)
				*vram++ = arg->items[self->value][i];
			break;
		}
		case VERB_COMMIT: {
			break;
		}
	}
	return ACT_NONE;
}
#pragma endregion
#pragma region methods
/**--------------------------------------------------------------------------
 **	page methods
 **------------------------------------------------------------------------*/
static void drawitem(LPCFGITEM item, int line, BOOL selected) {
	int row = line + 9;
	item->controller(item, selected ? VERB_PRINT_SEL : VERB_PRINT, &item->parent->parent->text[row][1]);
	for (int i = 1; i < 79; ++i)
		item->parent->parent->attr[row][i] = selected ? 0x22 : 0x07;
	if (selected) {
		memset(&item->parent->parent->text[7][1], SP, 78);
		sprintf(&item->parent->parent->text[7][1], " hint: %s", item->hint);
	}
}

static LPCFGITEM get_item(LPCFGPAGE self) {
	return &self->items[self->cur_idx];
}

static int call_current_item(LPCFGPAGE self, int verb, LPVOID arg) {
	LPCFGITEM item = self->get_item(self);
	return item->controller(item, verb, arg);
}

static void page_reflesh(LPCFGPAGE self) {
	for (int i = 0; i < self->item_cnt; ++i) {
		LPCFGITEM item = &self->items[i];
		item->controller(item, VERB_INIT, NULL);
		self->parent->drawitem(item, i, i == self->cur_idx);
	}
	self->parent->show(self->parent);
}

static void prepare_page(LPCFGPAGE self, LPVOID arg) {
	self->item_cnt = 0;
	self->cur_idx = 0;
	switch (self->page) {
		case 1: {
			LPDISKIMAGE image = (LPDISKIMAGE)arg;
			for (int r = 0; r < image->disk_count; ++r) {
				LPFLOPPYDISK disk = image->disks[r];
				LPCFGITEM item = &self->items[r];
				item->parent = self;
				item->arg = disk;
				self->item_cnt++;
				self->items[r].controller(&self->items[r], VERB_INIT, NULL);
				self->parent->drawitem(item, r, r == 0);
			}
			break;
		}
		default:
			for (int r = 0; self->items[r].type != TYPE_TERMINATOR; ++r) {
				self->items[r].parent = self;
				self->item_cnt++;
				if (self->items[r].type == TYPE_IMGSLOT) {
					int slot = (self->items[r].id & 0x0f) - 1;
					self->items[r].arg = n80device.fdu.mgr.slots[slot];
				}
				self->items[r].controller(&self->items[r], VERB_INIT, NULL);
				self->parent->drawitem(&self->items[r], r, r == 0);
			}
			break;
	}
}

static int handle_evt_page(LPCFGPAGE self, LPKEYEVT kv) {
	LPCFGITEM item = self->get_item(self);
	int action = item->controller(item, VERB_EVENT, kv);
	switch (self->page) {
		case 0:
			if (kv->key == VK_ESCAPE) {
				self->call(self, VERB_COMMIT, NULL);
				//n80lx.ini.flush(&n80lx.ini);
				return ACT_QUIT;
			}
			break;
		default: {
			if (action == ACT_QUIT || kv->key == VK_ESCAPE) {
				self->call(self, VERB_COMMIT, NULL);
				self->jmp_arg = NULL;
				self->jmp_page = 0;
				return ACT_CHANGEPAGE;
			}
			break;
		}
	}
	return action;
}
/**--------------------------------------------------------------------------
 **	config methods
 **------------------------------------------------------------------------*/
static void show(LPCONFIG self) {
	th->getRender()->postMessage(WM_REQ_DRAW_CUI, (WPARAM)self, NULL);
}
static int count_page(void) {
	int cnt = 0;
	for (int i = 0;; ++i) {
		if (page_items[i] == NULL)
			break;
		++cnt;
	}
	return cnt;
}
static void init_page(LPCONFIG self, int page, LPVOID arg) {
	self->page = page;
	self->maxpages = count_page();

	// init page objects
	for (int p = 0; p < self->maxpages; ++p) {
		self->pages[p].parent   = self;
		self->pages[p].page     = p;
		self->pages[p].items    = page_items[p];
		self->pages[p].get_item = get_item;
		self->pages[p].call     = call_current_item;
		self->pages[p].reflesh  = page_reflesh;
		self->pages[p].prepare  = prepare_page;
		self->pages[p].handler  = handle_evt_page;
	}
	// clear and draw base screen 
	for (int r = 0; r < 25; ++r) {
		for (int c = 0; c < 80; ++c) {
			BYTE atr = 0x23;
			if ((r == 4 || r == 5) && (c > 0 && c < 79))
				atr = 0x27;
			else if (r == 7 && (c > 0 && c < 79))
				atr = 0x26;
			else if ((r > 8 && r < 24) && (c > 0 && c < 79))
				atr = 0x07;
			self->attr[r][c] = atr;
			self->text[r][c] = SP;
		}
	}
	const char* title1 = "    \xea\xeaN80\xea\xea PC-8001 Emulator for DirectX";
	const char* title2 = "    Copyright 1993-2024 (C) H.Kanemoto, All Right Reserved.";
	const char* title3 = "System Config";
	const char* title4 = "usage:  [ESC] Quit/Back, [\x1e\x1f] Move, [SP] Toggle, [\x1d\x1c] Value";
	const char* title5 = "hint: ";
	memcpy(&self->text[1][2], title1, strlen(title1));
	memcpy(&self->text[2][2], title2, strlen(title2));
	memcpy(&self->text[4][2], title3, strlen(title3));
	memcpy(&self->text[5][2], title4, strlen(title4));
	memcpy(&self->text[7][2], title5, strlen(title5));

	self->pages[page].prepare(&self->pages[page], arg);
}

static BOOL dispatch(LPCONFIG self, LPKEYEVT kv) {
	LPCFGPAGE page = &self->pages[self->page];

	int action = page->handler(page, kv);
	if (action == ACT_DONE) {
		self->drawitem(page->get_item(page), page->cur_idx, true);
		self->show(self);
	}
	else if (action == ACT_CHANGEPAGE) {
		self->init_page(self, page->jmp_page, page->jmp_arg);
		self->show(self);
	}
	else if (action == ACT_QUIT) {
		return true;
	}
	else if ((kv->key == VK_TAB && !kv->SHIFT) || kv->key == VK_DOWN) { // [TAB] move to next item
		if (page->item_cnt > 1) {
			page->call(page, VERB_COMMIT, NULL);
			self->drawitem(page->get_item(page), page->cur_idx, false);
			if (++page->cur_idx >= page->item_cnt)
				page->cur_idx = 0;
			self->drawitem(page->get_item(page), page->cur_idx, true);
			self->show(self);
		}
	}
	else if ((kv->key == VK_TAB && kv->SHIFT) || kv->key == VK_UP) {
		if (page->item_cnt > 1) {
			page->call(page, VERB_COMMIT, NULL);
			self->drawitem(page->get_item(page), page->cur_idx, false);
			if (--page->cur_idx < 0)
				page->cur_idx = page->item_cnt - 1;
			self->drawitem(page->get_item(page), page->cur_idx, true);
			self->show(self);
		}
	}
	return false;
}

static void config(LPCONFIG self) {
	self->init_page(self, 0, NULL);
	self->show(self);
}
/**--------------------------------------------------------------------------
 **		Initializer
 **------------------------------------------------------------------------*/
export void init_config(LPCONFIG self) {
	self->init      = init_config;
	self->open      = config;
	self->init_page = init_page;
	self->dispatch  = dispatch;
	self->to_ascii  = to_ascii;
	self->drawitem  = drawitem;
	self->show      = show;
}
#pragma endregion
/**--------------------------------------------------------------------------
 **		_CAO֐
 **------------------------------------------------------------------------*/
LRESULT
n80_form::cfgWndproc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp) {
	switch (msg) {
	case WM_QUERYENDSESSION:		// Windows I𓾂悤ƂĂꍇB
	case WM_CLOSE:					// Window
	case WM_ENDSESSION:				// VXeI̓AvP[VI
	case WM_DESTROY:				// jꂽꍇAAvP[VI
		leave_config_mode();
		return 1;
	case  WM_PAINT: {
		PAINTSTRUCT	ps;
		HDC hdc = ::BeginPaint(mWnd, &ps);
		th->getRender()->postMessage(WM_REQ_DRAW_CUI, (WPARAM)this, NULL);
		::EndPaint(mWnd, &ps);
		return 0;
	}
	case WM_SYSKEYDOWN:
	case WM_KEYDOWN:
		n80win32.config.kv.key = (USHORT)(wp & 0xffff);
		n80win32.config.kv.ascii = n80win32.config.to_ascii(&n80win32.config.kv);
		if (wp == VK_SHIFT)
			n80win32.config.kv.SHIFT = TRUE;
		else if (n80win32.config.dispatch(&n80win32.config, &n80win32.config.kv)) {
			leave_config_mode();
		}
		break;
	case WM_SYSKEYUP:
	case WM_KEYUP:
		if (wp == VK_SHIFT)
			n80win32.config.kv.SHIFT = FALSE;
		break;
	case WM_SYSCOMMAND:
		break;
	default:
		return 1;
	}
	return 0;
}
/**--------------------------------------------------------------------------
 **		Config
 **------------------------------------------------------------------------*/
void
n80_form::enter_config_mode() {
	if (!m_config_mode && th->sync()) {
		m_config_mode = TRUE;
		dx->stop();
		ShowCursorFullScreen();
		::SetFocus(mWnd);
		// Z80syncŒ~ꂽRenderĊJ
		th->getRender()->desync();

		m_key_combo = KEYCOMB_CLEAR;

		n80win32.save_blt_context();
		init_config(&n80win32.config);
		n80win32.config.open(&n80win32.config);
	}
}
void
n80_form::leave_config_mode() {
	if (m_config_mode) {
		m_config_mode = FALSE;

		n80win32.restore_blt_context();
		n80win32.hard_volume = n80win32.get_hard_volume() - 1;
		n80win32.soft_volume = n80win32.get_soft_volume();
		saveProfile(TEXT("Default"));
		
		HideCursorFullScreen();
		::SetFocus(mWnd);
		n80device.attr.invalid();
		th->desync();
		dx->play();
	}
}
