/**==========================================================================

	PC-8001 Device Emulator for Win32 Ver 1.10
	Copyright (C)1993-2010 H.Kanemoto

	This is a free software. You may copy or distribute
	it without any permissions. But there are NO WARRANTY
	of use. Entire risk of use is with you.

	External assembler functions for IBM-PC with VC++ 5.0
	last modified  Apr,4,1998.

;==========================================================================*/
#include "pc-8001.h"
#ifdef _MSC_VER
#include <windows.h>
#include "../N80/N80.h"
#endif
#ifdef __cplusplus
extern "C"
{
#endif
#define export
extern void _blt_simd_level(int simd_level);
extern void blt8025(int line);
extern void blt8020(int line);
extern void blt4025(int line);
extern void blt4020(int line);
extern void blt8025x(int line);
extern void blt8020x(int line);
extern void blt4025x(int line);
extern void blt4020x(int line);
extern void blt8025v(int line);
extern void blt8020v(int line);
extern void blt4025v(int line);
extern void blt4020v(int line);
extern void bltblank(void);
extern void bltblankx(void);
extern void bltblankv(void);

typedef void (*PBLTFNC)(int);
typedef void (*PBLTBLK)(void);
// GRB to RGB order
static const BYTE	font_color[8]  = {0, 252, 250, 254, 249, 253, 251, 255};
static const BYTE	font_colord[8] = {0,   4,   2,   6,   1,   5,   3, 248};
static const DWORD	font_colorx[8] = {
	0x00000000, 0x000000FF, 0x0000FF00, 0x0000FFFF, 0x00FF0000, 0x00FF00FF, 0x00FFFF00, 0x00FFFFFF
};
static const DWORD	font_colorxd[8] = {
	0x00000000, 0x0000007F, 0x00007F00, 0x00007F7F, 0x007F0000, 0x007F007F, 0x007F7F00, 0x007F7F7F
};

static PBLTFNC	blt_funcs[] = {blt4020, blt4020x, blt4020v};
static PBLTFNC	fnc_tbl[]	= {blt4020, blt4025, blt8020, blt8025};
static PBLTFNC	fnc_tblx[]	= {blt4020x, blt4025x, blt8020x, blt8025x};
static PBLTFNC	fnc_tblv[]	= {blt4020v, blt4025v, blt8020v, blt8025v};
static PBLTBLK	blt_blks[]	= {bltblank, bltblankx, bltblankv};

static void chg_bltfunc(int screen_mode) {
	if (screen_mode >= 0 && screen_mode < 4) {
		blt_funcs[0] = fnc_tbl[screen_mode];
		blt_funcs[1] = fnc_tblx[screen_mode];
		blt_funcs[2] = fnc_tblv[screen_mode];
	}
}
static void _blt_set_simd_level(int simd_level) {
	_blt_simd_level(simd_level);
	if (n80win32.not_use_mmx)
		n80device.blt.bltavx = 0;
	else if (simd_level > 7 && n80win32.use_avx)
		n80device.blt.bltavx = 2;
	else
		n80device.blt.bltavx = 1;
}

static void _blt_font(int line, int bltavx) {
	blt_funcs[bltavx](line);
}

static void _blt_blank(int bltavx) {
	blt_blks[bltavx]();
}

static void __cdecl chg_palette(BOOL interlace) {
	n80win32.second_palettes   = interlace ? font_colord : font_color;
	n80win32.second_rgb_colors = interlace ? font_colorxd : font_colorx;
	for (int i = 0; i < 8; ++i) {
		n80device.blt.direct_palette_rgb_fix[0][i] = n80win32.prime_rgb_colors[i];
		n80device.blt.direct_palette_rgb_fix[1][i] = n80win32.second_rgb_colors[i];
	}
}

export void _n80_init_blt(void) {
	n80device.blt.chg_crt_mode	= chg_bltfunc;
	n80device.blt.font			= _blt_font;
	n80device.blt.blank			= _blt_blank;
	n80device.blt.chg_palette	= chg_palette;
	n80device.blt.set_simd_level= _blt_set_simd_level;
	n80win32.prime_palettes		= font_color;
	n80win32.prime_rgb_colors	= font_colorx;
	chg_palette(FALSE);
}

#ifdef __cplusplus
}	// extern "C"
#endif
